import asyncio
import inspect
import logging
from typing import Dict, Tuple, List, Optional, Union, Callable
from .rpc_stream import RPCStream
from . import msg_type as mtype
from .error import *

import msgpack

logger = logging.getLogger(__name__)


class RPCServer:

    def __init__(self, ws, handler, *, timeout=10, use_list=False):
        self.ws = ws
        self.timeout = timeout
        self.handler = handler
        self._packer = msgpack.Packer(use_bin_type=True)
        self._use_list = use_list
        self._tasks: Dict[int, Tuple[asyncio.Task, Optional[wsrpc.RPCStream]]] = {}

    async def run(self):
        async for data in self.ws:
            try:
                await self._on_data(data)
            except Exception as e:
                logger.exception(e)
        try:
            await asyncio.shield(self._join())
        except asyncio.CancelledError:
            await self._join()

    async def _join(self):
        for t,q in self._tasks.values():
            t.cancel()
        await asyncio.wait([t for t,q in self._tasks.values()], timeout=self.timeout)

    async def _on_data(self, data: bytes):
        msg = msgpack.unpackb(data, use_list=self._use_list)
        msgtype, msgid = msg[:2]

        if msgtype == mtype.REQUEST or msgtype == mtype.NOTIFY:
            method_name, params = msg[2:]

            method = getattr(self.handler, method_name, None)
            if method_name and method_name[0]!='_' and method:
                kwoa = inspect.getfullargspec(method).kwonlyargs
                if kwoa and kwoa[-1]=='request_stream':
                    q_size = getattr(self.handler, 'q_size', {}).get(method_name, getattr(self.handler, 'default_q_size', 0))
                    q = RPCStream(q_size)
                else:
                    q = None

                if inspect.isasyncgenfunction(method):
                    task = asyncio.create_task(self._on_request_gen(msgid, method, params, q))

                elif inspect.iscoroutinefunction(method):
                    task = asyncio.create_task(self._on_request(msgid, method, params, q))

                else:
                    try:
                        ret = self._call(method, params, None)
                    except Exception as e:
                        await self._send_error(msgid, str(e))
                    else:
                        await self._send_response(msgid, ret)
                    return

                self._tasks[msgid] = task, q
                task.add_done_callback(lambda f: self._tasks.pop(msgid, None))

            else:
                await self._send_error(msgid, f'{method_name} method not found.')

        elif msgtype == mtype.REQUEST_STREAM_CHUNCK:
            self._tasks[msgid][1].put_nowait(msg[2])

        elif msgtype == mtype.REQUEST_STREAM_END:
            self._tasks[msgid][1].put_nowait(StopAsyncIteration())

        elif msgtype == mtype.REQUEST_CANCEL:
            t = self._tasks.get(msgid)
            t and t[0].cancel()

        else:
            raise RPCServerError("unknown msgtype")

    def _call(self, method, params, q):
        ptype = type(params)
        if ptype == list or ptype == tuple:
            return method(*params, request_stream=q) if q else method(*params)
        elif ptype == dict:
            return method(**params, request_stream=q) if q else method(**params)
        else:
            raise Exception('Wrong parameters for '+ method.__name__)

    async def _on_request(self, msgid: int, method: Callable, params: Union[list, dict], q: Optional[RPCStream]) -> None:
        try:
            ret = await self._call(method, params, q)
        except (asyncio.CancelledError, Exception) as e:
            await self._send_error(msgid, str(e) or e.__class__.__name__)
        else:
            await self._send_response(msgid, ret)

    async def _on_request_gen(self, msgid: int, method: Callable, params: Union[list, dict], q: Optional[RPCStream]) -> None:
        try:
            async for resp in self._call(method, params, q):
                await self._send_stream_chunck(msgid, resp)
        except (asyncio.CancelledError, Exception) as e:
            await self._send_error(msgid, str(e) or e.__class__.__name__)
        else:
            await self._send_stream_end(msgid)


    async def _send_response(self, msgid: int, result) -> None:
        await self.ws.send(self._packer.pack((mtype.RESPONSE, msgid, None, result)))

    async def _send_error(self, msgid: int, err: str) -> None:
        await self.ws.send(self._packer.pack((mtype.RESPONSE, msgid, err, None)))

    async def _send_stream_chunck(self, msgid: int, chunck) -> None:
        await self.ws.send(self._packer.pack((mtype.RESPONSE_STREAM_CHUNCK, msgid, chunck)))

    async def _send_stream_end(self, msgid: int) -> None:
        await self.ws.send(self._packer.pack((mtype.RESPONSE_STREAM_END, msgid)))




import asyncio

class RPCStream(asyncio.Queue):
    def __init__(self, maxsize=0):
        asyncio.Queue.__init__(self, maxsize)

    def __aiter__(self):
        return self

    async def __anext__(self):
        d = await self.get()
        if isinstance(d, Exception):
            raise d
        else:
            return d

    async def close(self):
        await self.put(StopAsyncIteration())

    def force_put_nowait(self, o):
        # if queue is full, then pop out an element to make room
        self.full() and self.get_nowait()
        self.put_nowait(o)



import asyncio
import logging
import functools
from collections import Iterable
from . import msg_type as mtype
import msgpack
from .rpc_stream import RPCStream
from .error import *

logger = logging.getLogger(__name__)

class RPCFuture(asyncio.Future):

    def __init__(self, msgid, start, cancel, q_size):
        asyncio.Future.__init__(self)
        self._msgid = msgid
        self._start = start
        self._response_stream = None
        self._q_size = q_size
        self._cancel = cancel
        self._task = None
        self.add_done_callback(lambda f: self._task and self._task.cancel())

    @property
    def response_stream(self):
        if not self._response_stream:
            self._response_stream = RPCStream(self._q_size)
        return self._response_stream

    def cancel(self):
        if not self.done():
            self._response_stream and self._response_stream.force_put_nowait(RPCClientError('Cancelled by client.'))
            asyncio.create_task(self._cancel(self._msgid))
        return asyncio.Future.cancel(self)

    def __del__(self):
        self.cancel()

    def request(self):
        if not self._task:
            self._task = asyncio.create_task(self._start)

    def __aiter__(self):
        self.request()
        return self.response_stream

    def __await__(self):
        # await self._coro
        # self._coro.__await__()
        self.request()
        return asyncio.Future.__await__(self)


class RPCClient:
    def __init__(self, ws, *, use_list=False):
        self.ws = ws
        self._use_list = False
        self._packer = msgpack.Packer(use_bin_type=True)
        self._mid = 0
        self._tasks = {}
        asyncio.create_task(self._run())

    def _next_msgid(self):
        if self._mid >= 2**20:
            self._mid = 0
        self._mid += 1
        return self._mid

    async def _run(self):
        async for data in self.ws:
            try:
                msg = msgpack.unpackb(data, use_list=self._use_list)
                msgtype, msgid = msg[:2]
                if msgtype == mtype.RESPONSE:
                    err, result = msg[2:]
                    t = self._tasks.pop(msgid, None)
                    if t:
                        if err:
                            if not t.done():
                                e = RPCServerError(err)
                                t.set_exception(e)
                                t._response_stream and t._response_stream.force_put_nowait(e)
                        else:
                            t.set_result(result)

                elif msgtype == mtype.RESPONSE_STREAM_CHUNCK:
                    self._tasks[msgid].response_stream.put_nowait(msg[2])

                elif msgtype == mtype.RESPONSE_STREAM_END:
                    t = self._tasks.pop(msgid, None)
                    if t and not t.done():
                        t.response_stream.force_put_nowait(StopAsyncIteration())
                        t.set_result(None)

            except Exception as e:
                logger.exception(str(e))


    def __getattr__(self, method):
        return functools.partial(self._request, method)

    async def _send_request(self, msgid, method, params):
        await self.ws.send(self._packer.pack((mtype.REQUEST, msgid, method, params)))

    async def _send_stream_chunck(self, msgid, chunck):
        await self.ws.send(self._packer.pack((mtype.REQUEST_STREAM_CHUNCK, msgid, chunck)))

    async def _send_stream_end(self, msgid):
        await self.ws.send(self._packer.pack((mtype.REQUEST_STREAM_END, msgid)))

    async def _send_cancel(self, msgid):
        await self.ws.send(self._packer.pack((mtype.REQUEST_CANCEL, msgid)))

    async def _req_iter(self, msgid, iter):
        if isinstance(iter, Iterable):
            for i in iter:
                await self._send_stream_chunck(msgid, i)
        else:
            async for i in iter:
                await self._send_stream_chunck(msgid, i)
        await self._send_stream_end(msgid)

    def _request(self, method, *args, **kwargs):
        msgid = self._next_msgid()
        req_iter = kwargs.pop('request_stream', None)
        async def start():
            await self._send_request(msgid, method, args or kwargs)
            if req_iter:
                await self._req_iter(msgid, req_iter)

        fut = RPCFuture(msgid=msgid, start=start(), cancel=self._send_cancel, q_size=kwargs.pop('q_size', 0))
        self._tasks[msgid] = fut
        return fut



import setuptools

with open("./README.md", 'r') as f:
    long_description = f.read()

with open('./requirements.txt', 'r') as f:
    requirements = [a.strip() for a in f]

setuptools.setup(
    name="wsmprpc",
    version="0.2.1",
    author="Huang Yan",
    author_email="hyansuper@foxmail.com",
    description="python msgpack RPC over websocket",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hyansuper/wsmprpc",
    packages=['wsmprpc'],
    install_requires=requirements,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)